#include "Reaper.h"

inline size_t ClearBits(size_t &datum, size_t bits) {
    size_t temp = datum | bits;
    return temp ^ bits;
}

inline size_t * MakeAddress1(size_t &field) {
    return (size_t *) ClearBits(field, MSB | 3);
}

inline CONS_cell * MakeAddress2(size_t &field) {
    return (CONS_cell *) ClearBits(field, MSB | 3);
}

// An exception to the rules is being allowed, where
// a pointer to CONS_cell is not aligned!
inline CONS_cell * MakeList(size_t &field) {
    return (CONS_cell *) ClearBits(field, MSB | 2);
}

inline bool DetermineWhite(CONS_cell &cell) {
    return (! (cell.CAR & MSB));
}

inline bool DetermineWhite(size_t &field) {
    //  For Immediate type Address.
    return (! (field & MSB));
}

// An exception to the rules is being tested for, where
// a field corresponding to a CONS_cell is not aligned!
inline bool DetermineList(size_t &field) {
    return ((field & 3) == 1);
}


inline int ReapBlock(CONS_cell *block_base) {
    int i = (BLOCK_SIZE / sizeof(CONS_cell)) - 1;
    bool isFree = false;
    int count = 0;

#ifdef PF_CALLED

    size_t last = (size_t) block_base;	// !

    /*	If the function PreserveFree() is Not to be called
     * by the GC, then the preceding line Must be changed to: */

#else
    size_t last = 0;
#endif

    for ( ; i > 0 ; i--) {
        isFree = DetermineWhite(block_base[i]);
        if (isFree) {
            block_base[i].CDR = last;
            block_base[i].CAR = 0;

            last = ((size_t) (block_base + i)) | 1;
            ++count;
        }
    }

    block_base->CAR = last;

    return count;
}

inline int AcquireReapBlock(CONS_cell *&here, ChoiceMutex *m) {
    size_t field = (size_t) here;
    int count = 0;

    if (! field) return 0;

    m->Reserve(field);
    m->Reserve(field + BLOCK_SIZE);

    count = ReapBlock(here) + 1;
    here = MakeAddress2(here->CDR);

    m->Release(field);
    m->Release(field + BLOCK_SIZE);
    return count;
}

long int ReapLists(CONS_cell *root, ChoiceMutex *m) {
    CONS_cell *ptr = root;
    long int count = 0;
    int temp = 0;

    while ( (temp = AcquireReapBlock(ptr, m)) ) {
        count += (temp - 1);
    }
    return count;
}

int FreeNonImmediate1(CONS_cell *root) {
    bool flag = (root != 0);
    size_t *ptr = NULL;
    size_t temp = 0;
    int count = 0;

    while (flag) {
        temp = ClearBits(root->CAR, MSB);
        ptr = (size_t *) temp;

        /*	The risk is very high here that free() could get called
         * on something that was never malloc'ed. To reduce that risk,
         * I'm going to say for the moment that, non-immediate
         * objects which were malloc'ed, never had any bits in their
         * first word set, aside from the first byte. It must at least
         * be a 32-bit word, so that the chances of that happening
         * for any other reason are quite low...
         *
         * 	If erroneously pointed to as immediate type Address,
         * FIXUMs from 0 to +31 would apply, as would certain
         * (non-printing) characters... */

        if (temp && (! (temp & 3)) && (! (*ptr & MSB)) &&
                ((*ptr & 0xFF) < 128)) {


//		if (temp && (! (temp & 3)) && (! (*ptr & MSB)) &&
//				(*ptr < 128)) {

            // Taking on faith here, that this won't blow up...
            free((void *) ptr);
            root->CAR = 0;
            ++count;
        }

        temp = root->CDR;
        flag = ((temp & 3) == 1);
        //  Establish CDR is a list.

        if (flag) root = MakeAddress2(temp);
    }
    return count;
}


/* An exception to the rules is being created below, where
 * a field corresponding to CONS_cell is not aligned!
 * But, after execution, *output will be aligned.
 * *input will also be aligned, but hold nothing of further value.
 * Also, this function orphans cells, which will only be
 * recycled, during the next GC cycle.
 *
 */

void RemoveNILs(CONS_cell *&input, CONS_cell *&output) {
    CONS_cell *temp = NULL;
    size_t modify = ((size_t) input) | MSB | 1;
    modify ^= MSB;

    if (! input) return;

    while (DetermineList(modify)) {
        modify ^=  1;	// A valid address again, synonym of input.
        temp = MakeList(((CONS_cell *) modify)->CDR);
        if (((CONS_cell *) modify)->CAR) {
            if (output) {
                ((CONS_cell *) modify)->CDR = ((size_t) output) | 1;
            } else {
                ((CONS_cell *) modify)->CDR = 0;
            }
            output = (CONS_cell *) modify;
        }
        modify = (size_t) temp;
    }
}

// If the interpreter wants to allocate using malloc(), after doing so,
// and before editing this list, it must acquire the Mutex on
// resource (0).
int FreeNonImmediate(CONS_cell *&root, ChoiceMutex *m) {
    CONS_cell *temp_out = NULL;
    int count = 0;

    m->Reserve(0);

    count = FreeNonImmediate1(root);
    RemoveNILs(root, temp_out);
    root = temp_out;

    m->Release(0);

    return count;
}
