#include "Tricolor_Map.h"


INLINE bool DetermineGray(CONS_cell &cell) {
    return ((cell.CAR & MSB) && (! (cell.CDR & MSB)));
}

INLINE bool DetermineWhite(CONS_cell &cell) {
    return (! (cell.CAR & MSB));
}

INLINE bool DetermineWhite(size_t &field) {
    //  For Immediate type Address.
    return (! (field & MSB));
}

INLINE size_t ClearBits(size_t &datum, size_t bits) {
    size_t temp = datum | bits;
    return temp ^ bits;
}

INLINE void Mark_cell_gray(CONS_cell &cell, ChoiceMutex *m) {
    size_t here = (size_t) &(cell.CAR);

    if (m) m->Reserve(here);

    cell.CAR |= MSB;

    if (m) m->Release(here);

}

INLINE void Mark_cell_black(CONS_cell &cell, ChoiceMutex *m) {
    size_t here = (size_t) &(cell.CDR);

    if (m) m->Reserve(here);

    cell.CDR |= MSB;

    if (m) m->Release(here);

}

INLINE void Mark_field(size_t &field) {
    field |= MSB;
}

INLINE void UnMark_field(size_t &field) {
    size_t temp = field & MSB;
    field ^= temp;
}

INLINE size_t * MakeAddress1(size_t &field) {
    return (size_t *) ClearBits(field, MSB | 3);
}

INLINE CONS_cell * MakeAddress2(size_t &field) {
    return (CONS_cell *) ClearBits(field, MSB | 3);
}

INLINE CONS_cell * MakeList(size_t &field) {
    return (CONS_cell *) ClearBits(field, MSB | 2);
}


/*	A concept which the following 2, critical functions embrace is,
 * that LSBs '11' state immediate data-type Address, pointing to
 * a struct, the first word of which determines the non-immediate data-
 * type. A further subdivision of these data-types is, that if the
 * first byte is less than 128, they have been allocated via malloc().
 * However, if the first byte is greater than 127, they have been
 * allocated in a CONS block, And, their second word is actually
 * a List for the internal representation. In that case, the LSBs
 * of that first byte must be '11', so that the first word
 * will not be traversed by the mapping algorithm. Thus, those last
 * non-immediate types also need to have a (dereferenced) first
 * byte value which is a mutliple of 4, minus 1.
 *
 */

void EnterCellMappingL2(CONS_cell &cell, ChoiceMutex *m) {
    Mark_cell_gray(cell, m);
    Mark_cell_black(cell, m);

    size_t temp1 = cell.CAR & 3;

    if (! temp1) {
        size_t *address = MakeAddress1(cell.CAR);
        if (address && ((*address & 0xFF) < 128)) {
            //  Immediate type = Address,
            //  malloc'ed .
            Mark_field(*address);
        } else if (address) {
            //  Immediate type = Address,
            //  allocated as CONS-cell.
            temp1 ^= 1;
        }
    }

    temp1 ^= 1;
    if (! temp1) {
        //  Immediate type = List.
        CONS_cell * temp2 = MakeAddress2(cell.CAR);
        if (temp2 && DetermineWhite(*temp2)) {
            Mark_cell_gray(*temp2, m);
        }
    }

    temp1 = cell.CDR & 3;

    if (! temp1) {
        size_t *address = MakeAddress1(cell.CDR);
        if (address && ((*address & 0xFF) < 128)) {
            //  Immediate type = Address,
            //  malloc'ed .
            Mark_field(*address);
        } else if (address) {
            //  Immediate type = Address,
            //  allocated as CONS-cell.
            temp1 ^= 1;
        }
    }

    temp1 ^= 1;
    if (! temp1) {
        //  Immediate type = List.
        CONS_cell * temp2 = MakeAddress2(cell.CDR);
        if (temp2 && DetermineWhite(*temp2)) {
            Mark_cell_gray(*temp2, m);
        }
    }
}

void EnterCellMapping(CONS_cell &cell, ChoiceMutex *m) {
    Mark_cell_gray(cell, m);
    Mark_cell_black(cell, m);

    size_t temp1 = cell.CAR & 3;

    if (! temp1) {
        size_t *address = MakeAddress1(cell.CAR);
        if (address && ((*address & 0xFF) < 128)) {
            //  Immediate type = Address,
            //  malloc'ed .
            Mark_field(*address);
        } else if (address) {
            //  Immediate type = Address,
            //  allocated as CONS-cell.
            temp1 ^= 1;
        }
    }

    temp1 ^= 1;
    if (! temp1) {
        //  Immediate type = List.
        CONS_cell * temp2 = MakeAddress2(cell.CAR);
        if (temp2 && DetermineWhite(*temp2)) {
            EnterCellMappingL2(*temp2, m);
        }
    }

    temp1 = cell.CDR & 3;

    if (! temp1) {
        size_t *address = MakeAddress1(cell.CDR);
        if (address && ((*address & 0xFF) < 128)) {
            //  Immediate type = Address,
            //  malloc'ed .
            Mark_field(*address);
        } else if (address) {
            //  Immediate type = Address,
            //  allocated as CONS-cell.
            temp1 ^= 1;
        }
    }

    temp1 ^= 1;
    if (! temp1) {
        //  Immediate type = List.
        CONS_cell * temp2 = MakeAddress2(cell.CDR);
        if (temp2 && DetermineWhite(*temp2)) {
            EnterCellMappingL2(*temp2, m);
        }
    }
}

// Here it's understood, that the address doesn't really point
// to a CONS object...
EXPORT_INLINE void Mark_address_gray(CONS_cell *address,
                                     ChoiceMutex *m) {
    size_t temp = 0;
    if (! address) return;
    temp = address->CAR;

    if ( ((temp & 0xFF) > 128 ) && ((temp & 2) > 1) ) {
        // First byte >= 128 . Therefore, allocated in a CONS block.
        // *CAR must not be navigable.

        EnterCellMapping(*address, m);

    } else {
        // First byte < 128. Therefore, *CDR is not navigable.

        Mark_field(temp);
    }
}

INLINE bool Map_cell(CONS_cell &cell, ChoiceMutex *m) {
    if (DetermineGray(cell)) {
        EnterCellMapping(cell, m);
        return true;
    } else {
        return false;
    }
}

INLINE bool GC_Scan1(CONS_cell *ptr, ChoiceMutex *m) {
    int i = 1;
    int cell_count = BLOCK_SIZE / sizeof(CONS_cell);
    bool was_gray = false;

    for ( ; i < cell_count ; i++) {
        ptr++;
        was_gray = (Map_cell(*ptr, m) || was_gray);
    }
    return was_gray;
}

bool GC_Scan(CONS_cell *block_base, ChoiceMutex *m) {
    bool temp = false;
    while (GC_Scan1(block_base, m)) {
        temp = true;
    }
    return temp;
}


/*	The next assumption is, that a LISP system has a number of
 * blocks, of CONS-cells, the first cell of which forms a singly-
 * linked list joining them all. The CDR points to the next
 * block-base, while the CAR will branch out to the individual
 * free-cell sub-list.
 *
 * 	The CONS-cell at the base of the blocks differs from ones
 * used by LISP, in that these do not have any special bits.
 * Their fields are just addresses, and the CDR of the last
 * block-base is NULL.
 *
 * 	The following function will purge each block of gray cells,
 * until the entire set of blocks is left with none of them.
 *
 * 	Hint: Purging one block of gray cells will often cause
 * cells to be marked gray in the other blocks.
 *
 */

void MapBlockList(CONS_cell *first_block, ChoiceMutex *m) {
    CONS_cell *current_block = first_block;
    CONS_cell *next_block = NULL;
    bool still_dirty = true;

    while (still_dirty) {
        still_dirty = false;
        while (current_block) {
            next_block = MakeAddress2(current_block->CDR);
            still_dirty = (GC_Scan(current_block, m) || still_dirty);
            current_block = next_block;
        }
        current_block = first_block;
        next_block = NULL;
    }
}


/*	The next assumption is, that under the subject of mapping,
 * certain Lists need to be preserved, but are not given as root
 * for other cells to be kept alive. Their CAR fields are to be
 * ignored.
 *
 * 	They will be marked 'black'.
 *
 * 	Because these lists are fully part of the LISP state, they have
 * the tag fields, and variability in their LSB.
 *
 */

void PreserveMapList(CONS_cell *root, ChoiceMutex *m) {
    bool flag = (root != 0);
    size_t temp = 0;

    while (flag) {

        Mark_cell_gray(*root, m);
        Mark_cell_black(*root, m);

        temp = root->CDR;

        flag = ((temp & 3) == 1);
        //  Establish CDR is a list.

        if (flag) root = MakeAddress2(temp);
    }
}

void UnMarkCARs(CONS_cell *root) {
    bool flag = (root != 0);
    size_t *address = NULL;
    size_t temp = 0;

    while (flag) {
        temp = root->CAR;
        address = (size_t *) temp;
        if ((! (temp & 3)) && ((*address & 0xFF) < 128)) {
            /*  Establish it's a non-immediate data-type.
             * The assumption is being made here, that fields
             * the LSBs of which are '00', have the immediate
             * data-type Address.
             * 	If the first byte that address points to is
             * less than 128, we'll say it was malloc'ed. */

            UnMark_field(*address);
        }

        temp = root->CDR;
        flag = ((temp & 3) == 1);
        //  Establish CDR is a list.

        if (flag) root = MakeAddress2(temp);
    }
}

/*	In a way that exceeds what PreserveMapList() does, there
 * also exists a needed function, from which the CAR fields of a
 * list are Not Ignored, but rather, all marked gray, just so that
 * LISP's various stacks can be marked gray efficiently...
 *
 * 	Theoretically, just marking the first cell of such a list
 * gray, should accomplish the same thing eventually. But,
 * that would only proceed at the rate of 2 cells / pass. In some
 * cases, an entire list must be marked gray quickly...
 *
 */

void Mark_list_gray(CONS_cell *list_root, ChoiceMutex *m) {
    CONS_cell *ptr = NULL;
    bool flag = (list_root != 0);
    size_t temp = 0;

    while (flag) {

        // Cells which were once black, will date back to an earlier
        // GC cycle, as one gets far enough into the list...
        Mark_cell_gray(*list_root, m);
        Mark_cell_black(*list_root, m);

        temp = list_root->CAR;
        ptr = MakeAddress2(temp);

        if ((temp & 3) == 1) EnterCellMapping(*ptr, m);
        // Immediate type = List.

        if (! (temp & 3) ) Mark_address_gray(ptr, m);
        // Immediate type = Address or NIL.

        temp = list_root->CDR;
        ptr = MakeAddress2(temp);

        flag = ((temp & 3) == 1) && (DetermineWhite(*ptr));
        // Establish CDR is a list.
        // Avoid circularity.

        if (flag) list_root = ptr;
    }

}

INLINE bool InitBlockToWhite(CONS_cell *&root, ChoiceMutex *m) {
    int i = (BLOCK_SIZE / sizeof(CONS_cell)) - 1;
    size_t field = ((size_t) root) | MSB | 3;
    size_t field2 = 0;
    field ^= MSB | 3;

    if (! field) return false;

    m->Reserve(field);
    m->Reserve(field + BLOCK_SIZE);

    for ( ; i > 0 ; i--) {
        field = root[i].CAR & MSB;
        root[i].CAR ^= field;

        field = root[i].CDR & MSB;
        root[i].CDR ^= field;
    }

    field2 = root->CDR;

    m->Release(field);
    m->Release(field + BLOCK_SIZE);

    root = MakeAddress2(field2);

    return true;
}

void InitListsToWhite(CONS_cell *root, ChoiceMutex *m) {
    while (InitBlockToWhite(root, m)) { }
}


// Finally, a function must be called early, to preserve the
// lists of free CONS-cells.

void PreserveFree(CONS_cell *root, ChoiceMutex *m) {
    size_t temp = (size_t) root;
    size_t present = 0;
    size_t next = 0;

    while (temp) {

        present = temp;
        next = present + BLOCK_SIZE;

        m->Reserve(present);
        m->Reserve(next);

        temp = root->CAR;
        if (temp) PreserveMapList(MakeAddress2(temp), NULL);
        temp = root->CDR;

        m->Release(present);
        m->Release(next);

        if (temp) root = MakeAddress2(temp);
    }
}
